import numpy as np
import h5py


def find_dataset(h5_file): 
    '''Automatically find the dataset name in the h5 file containing the input movie.

    Inputs: 
        h5_file (h5py.File): the input h5 file containing the movie.

    Outputs:
        dset (str): the dataset name in the h5 file containing the input movie.
    '''
    list_dset = list(h5_file.keys())
    num_dset = len(list_dset)
    if num_dset == 1:
        dset = list_dset[0]
    else:
        list_size = np.zeros(num_dset, dtype='uint64')
        for (ind,dset) in enumerate(list_dset):
            if isinstance(h5_file[dset], h5py.Dataset):
                list_size[ind] = h5_file[dset].size
        ind_dset = list_size.argmax()
        if list_size[ind_dset] > 0:
            dset = list_dset[ind_dset]
        else:
            raise(KeyError('Cannot find any dataset. Please do not put the movie under any group'))
    return dset




import numpy as np
import multiprocessing as mp
from scipy.io import savemat
from nmfunmix1_pertmin_res_MSE_novideo import nmfunmix1


def use_nmfunmix(traces, bgtraces, outtraces, list_neighbors, list_alpha=[0], Qclip=0, \
        th_pertmin=1, epsilon=0, th_residual=0, nbin=1, bin_option='downsample', flexible_alpha=True):
    ''' Unmix the traces of all neurons using NMF, and obtain the unmixed traces and the mixing matrix. 
    Inputs: 
        traces (numpy.ndarray of float, shape = (T,n)): The raw traces of all neurons.
            Each column represents a trace of a neuron, which can be a trace of F, dF/F, or SNR.
        bgtraces (numpy.ndarray of float, shape = (T,n)): All the raw background traces.
            Each column represents a background trace corresponding to a neuron.
        outtraces (numpy.ndarray of float, shape = (T,n)): The raw traces of all the outside activities.
            Each column represents an outside trace corresponding to a neuron.
        list_neighbors (list of list of int): 
            Each element is a list of indeces of neighboring neurons of a neuron.
        list_alpha (list of float, default to [0]): A list of alpha to be tested.
            The elements should be sorted in ascending order.
        Qclip (float, default to 0): Traces lower than this quantile are clipped to this quantile value.
            Qclip = 0 means no clipping is applied. 
        th_pertmin (float, default to 1): Maximum pertentage of unmixed traces equaling to the trace minimum.
            th_pertmin = 1 means no requirement is applied. 
        epsilon (float, default to 0): The minimum value of the input traces after scaling and shifting. 
        th_residual (float, default to 0): If not zero, The redisual of unmixing should be smaller than this value.
        nbin (int, default to 1): The temporal downsampling ratio.
            nbin = 1 means temporal downsampling is not used.
        bin_option (str, can be 'downsample' (default), 'sum', or 'mean'): 
            The method of temporal downsampling. 
            'downsample' means keep one frame and discard "nbin" - 1 frames for every "nbin" frames.
            'sum' means each binned frame is the sum of continuous "nbin" frames.
            'mean' means each binned frame is the mean of continuous "nbin" frames.
        flexible_alpha (bool, default to True): Whether a flexible alpha strategy is used 
            when the smallest alpha in "list_alpha" already caused over-regularization.
            False means the final alpha is the smallest element in "list_alpha".
            True means trying to recursively divide the smallest alpha by 2 until no over-regularization exists.

    Outputs:
        demix (numpy.ndarray of float, shape = (T,n)): The resulting unmixed traces. 
            Each column is the unmixed trace of a neuron.
        list_mixout (list of numpy.ndarray of float, shape = (n1,n1)): 
            Each element is the row-normalized mixing matrix for the NMF of each neuron.
        list_alpha_final (list of float): Each element is the final chosen alpha for the NMF of each neuron. 
            It might be one of the elements in "list_alpha", or a value smaller than the first element.
        list_MSE (list of numpy.ndarray of float, shape = (n1,)): 
            Each element is the mean squared error (NMF residual) between 
            the input traces and the NMF-reconstructed traces for the NMF of each neuron.
        list_n_iter (list of int): Each element is the number of iterations 
            to achieve NMF convergence for the NMF of each neuron.
    '''

    n = traces.shape[1]
    # Background subtraction for neuron and outside traces
    traces_clip = traces-bgtraces
    outtrace_clip = outtraces-bgtraces
    if Qclip>0: # Clip the bottom "Qclip" quantile, to eliminate negative transients.
        np.clip(traces_clip, np.quantile(traces_clip, Qclip, axis=0), None, out=traces_clip)
        np.clip(outtrace_clip, np.quantile(outtrace_clip, Qclip, axis=0), None, out=outtrace_clip)

    # results = []
    # for i in range(n):
    #     result = nmfunmix1(i, traces_clip[:, list_neighbors[i]], bgtraces[:,i], list_alpha, \
    #         th_pertmin, epsilon, th_residual, nbin, bin_option)
    #     results.append(result)

    # Apply NMF to unmix each group of input traces corresponding to each neuron.
    p = mp.Pool(mp.cpu_count())
    results = p.starmap(nmfunmix1, [(i, traces_clip[:, list_neighbors[i]], outtrace_clip[:,i], list_alpha, 
        th_pertmin, epsilon, th_residual, nbin, bin_option, flexible_alpha) for i in range(n)], chunksize=1)
    p.close()

    # See the explanation of "nmfunmix1" for detailed explanation of these output quantities
    list_traceout = ([x[0].T for x in results])
    demix = np.concatenate([x[0:1] for x in list_traceout], 0).T

    list_mixout = ([x[1] for x in results])
    list_final_alpha = np.array([x[5] for x in results])
    list_MSE = [x[6] for x in results]
    list_n_iter = [x[8] for x in results]

    # list_outtrace = ([x[2] for x in results])
    # list_tempmixIDs = ([x[3] for x in results])
    # list_subtraces = ([x[4] for x in results])
    # list_tracein = ([x[7].T for x in results])

    # Can uncomment the previuos and following lines to save the unmixing results to a mat file,
        # mainly for debugging perpose.
    # savemat('demixtest_{}.mat'.format(alpha[0]), {'demix':demix, 'list_tracein':list_tracein, 'list_traceout':list_traceout,\
    #     'list_mixout':list_mixout, 'list_neighbors':list_neighbors, 'list_subtraces':list_subtraces, \
    #     'list_outtrace':list_outtrace, 'list_final_alpha':list_final_alpha, 'list_tempmixIDs':list_tempmixIDs})

    return demix, list_mixout, list_MSE, list_final_alpha, list_n_iter



from scipy import special
import numpy as np
from sklearn.decomposition import NMF
from sklearn.exceptions import ConvergenceWarning
import warnings
warnings.simplefilter('ignore', ConvergenceWarning)
np.seterr(divide='ignore',invalid='ignore')
from sklearn.metrics import mean_squared_error

from bin_trace_video import bin_trace


def nmfunmix(Ftmix, nbin=1, tol=1e-4, max_iter=20000, alpha=1, l1_ratio=0.5, epsilon=0, bin_option='downsample'):
    ''' Unmix the input traces in Ftmix using NMF, and obtain the unmixed traces and the mixing matrix. 
    Inputs: 
        Ftmix (numpy.ndarray of float, shape = (T,n)): The input mixed traces.
            Each column represents an input temporal trace, which can be a trace of F, dF/F, or SNR.
            The first column (Ftmix[:,0]) is the time series of the neuron of interest.
            The last column (Ftmix[:,-1]) should be the time series of the outside activities
            - the outside activities should be the average of the pixels around the neuron of interest, 
            and these pixels should not be in the masks of the neighboring neuorns included in "Ftmix".
            Each of the remaining columns (Ftmix[:,1:-1]), if any, should be the time series of a neighboring neuron.
        nbin (int, default to 1): The temporal downsampling ratio.
            nbin = 1 means temporal downsampling is not used.
        tol (float, default to 1e-4): Tolerance of the stopping condition in NMF.
        max_iter (int, default to 20000): Maximum number of iterations before timing out in NMF.
        alpha (float, default to 0): The regularization parameter, 
            or the weight of the regularization term in the cost function. 
            alpha = 0 means no regularization, and should lead to a trivial solution (traceout = Ftmix).
            Larger alpha means higher regularization, which can cause shorter processing time 
            and more seperated signals, but if the alpha is too large, it can cause merged activity 
            from multiple neurons, a flat baseline, or even an output of identically zero.
        l1_ratio (float, 0 <= l1_ratio <= 1, default to 0.5): 
            The ratio controling the weight between the Frobenius norm and the L1 norm. 
            The panelty term is (1-l1_ratio) * Frobenius_norm + l1_ratio * L1_norm.
            For l1_ratio = 0, the penalty is a pure Frobenius norm.
            For l1_ratio = 1, the penalty is a pure L1 norm.
            For 0 < l1_ratio < 1, the penalty is a combination of L1 and Frobenius norms.
        epsilon (float, default to 0): The minimum value of the input traces after scaling and shifting. 
        bin_option (str, can be 'downsample' (default), 'sum', or 'mean'): 
            The method of temporal downsampling. 
            'downsample' means keep one frame and discard "nbin" - 1 frames for every "nbin" frames.
            'sum' means each binned frame is the sum of continuous "nbin" frames.
            'mean' means each binned frame is the mean of continuous "nbin" frames.

    Outputs:
        traceout (numpy.ndarray of float, shape = (T,n)): The resulting unmixed traces. 
            Each column is the unmixed trace corresponding to each column in Ftmix.
        mixout (numpy.ndarray of float, shape = (n,n)): The row-normalized mixing matrix.
        tempmixIDs (numpy.ndarray of float, binary value, shape = (n,n)):
            The assistent transformation matrix used to match raw output traces to the input traces.
        subtraces (list of int): The index of identically zero output traces.
            Idealy should be empty.
        MSE (numpy.ndarray of float, shape = (n,)): 
            Mean squared error (NMF residual) between the input traces and the NMF-reconstructed traces.
        n_iter (int): Number of iterations to achieve NMF convergence.

    Note:
        The matrices in the code are often the transposes of the corresponding matrices in the papre.
            Ftmix is the transpose of F_meas,
            Ftdemix is the transpose of F_sep,
            mixout is the transpose of M,
            tempmixIDs is the transpose of P.

    '''

    (T, n) = Ftmix.shape # T is the number of frames, n is the number of components

    # Normalize and shift the traces
    medFt = np.median(Ftmix, 0)
    minFt = Ftmix.min() 
    Ftmix = Ftmix - minFt
    q12 = np.quantile(Ftmix[:,0], [0.25, 0.5], axis=0) 
    noise = (q12[1]-q12[0])/(np.sqrt(2)*special.erfinv(0.5))
    Ftmix = Ftmix/noise + epsilon

    # This is the NMF call
    if nbin == 1: # not use temporal downsampling
        Ftmix_bin = Ftmix
    else: # use temporal downsampling
        Ftmix_bin = bin_trace(Ftmix, nbin, bin_option)

    # Initialize NMF
    if n <= min(Ftmix_bin.shape):
        init = 'nndsvdar'
    else:
        init = 'random'
    nmf = NMF(init=init, n_components=n, tol=tol, max_iter=max_iter,
              alpha=alpha, l1_ratio=l1_ratio)
    
    # NMF fit
    Ftdemix_bin = nmf.fit_transform(Ftmix_bin)
    mixout = nmf.components_
    n_iter = nmf.n_iter_
    if np.any(mixout):
        if nbin == 1:
            Ftdemix = Ftdemix_bin
        else: 
            # if using temporal downsampling, use the fitted mixing matrix 
            # to unmixed the input traces in the origianl temporal resolution
            Ftdemix = nmf.transform(Ftmix)
        Ftmix_guess = nmf.inverse_transform(Ftdemix) # NMF-reconstructed traces
        # Ftmix ~= Ftmix_guess = Ftdemix.dot(mixout)
    else: # Extreme over-regularization caused all output traces to be identically zero
        print('NMF not found. Please try with lower alpha.')
        traceout = np.zeros_like(Ftmix)
        mixout = np.zeros((n, n))
        tempmixIDs = np.zeros((n, n))
        subtraces = np.arange(n)
        MSE = np.zeros((n), dtype = Ftmix.dtype)
        return traceout, mixout, tempmixIDs, subtraces, MSE, n_iter

    # Assign output traces to input traces
    mixout_sum = mixout.sum(1)[:,np.newaxis] 
    num_outtraces = mixout_sum.nonzero()[0].size

    Ftdemix = Ftdemix * mixout_sum.T
    mixout = mixout / mixout_sum  # normalize mixout by each row
    mixout[np.isnan(mixout)] = 0

    # Iteratively assgin outputs according to the maximum element of normalized mixout
    tempmixIDs = np.zeros((n, n))
    tempmixout = mixout.copy()
    for _ in range(num_outtraces):
        # indmax = np.nanargmax(np.abs(tempmixout))
        # indmax1 = indmax // n
        # indmax2 = indmax % n
        if np.all(np.isnan(tempmixout)):
            continue
        (indmax1, indmax2) = (np.nanmax(tempmixout)==tempmixout).nonzero()
        tempmixIDs[indmax1, indmax2] = 1
        tempmixout[indmax1, : ] = 0
        tempmixout[: , indmax2] = 0
        tempmixout = tempmixout / tempmixout.sum(1)[:,np.newaxis]

    # reorder and scale the outputs to match the inputs
    traceout = Ftdemix.dot(tempmixIDs * mixout)
    mixout = tempmixIDs.T.dot(mixout)  # reorder the mixing matrix
    # mixout = mixout / np.diag(mixout)[:,np.newaxis]
    subtraces = (traceout.sum(0) == 0).nonzero()[0]
    if subtraces.size:
        # If any output trace is identically zero, use the NMF residual as the output trace,
        # which is more informative than identically zero. 
        traceout[:, subtraces] = Ftmix[:, subtraces] - Ftmix_guess[:, subtraces]
    traceout = (traceout-epsilon) * noise + minFt  # Rescale the output traces to the actual inputs
    # Align the medians of the outputs to the inputs
    traceout = traceout - np.median(traceout, 0) + medFt
    MSE = mean_squared_error(Ftmix, Ftmix_guess, multioutput='raw_values')

    return traceout, mixout, tempmixIDs, subtraces, MSE, n_iter



import os
import sys
import time
import numpy as np
from scipy.io import savemat, loadmat
import h5py
from utils import find_dataset

if (sys.version_info.major+sys.version_info.minor/10)>=3.8:
    from multiprocessing.shared_memory import SharedMemory
    from traces_from_masks_mp_shm_neighbors import traces_bgtraces_from_masks_shm_neighbors
else:
    from traces_from_masks_mp_mmap_fn_neighbors import traces_bgtraces_from_masks_mmap_neighbors
from traces_from_masks_numba_neighbors import traces_bgtraces_from_masks_numba_neighbors
from use_nmfunmix_mp_MSE_novideo import use_nmfunmix


def run_TUnCaT(Exp_ID, filename_video, filename_masks, dir_traces, list_alpha=[0], Qclip=0, \
        th_pertmin=1, epsilon=0, th_residual=False, nbin=1, \
        bin_option='downsample', multi_alpha=True, flexible_alpha=True):
    ''' Unmix the traces of all neurons in a video, and obtain the unmixed traces and the mixing matrix. 
        The video is stored in "filename_video", and the neuron masks are stored in "filename_masks".
        The output traces will be stored in "dir_traces".
    Inputs: 
        Exp_ID (str): The name of the video.
        filename_video (str): The file path (including file name) of the video.
            The video file must be a ".h5" file. 
        filename_masks (str): The file path (including file name) of the neuron masks.
            The file must be a ".mat" file, and the masks are saved as variable "FinalMasks". 
        dir_traces (str): The folder to save the unmixed traces.
        list_alpha (list of float, default to [0]): A list of alpha to be tested.
            The elements should be sorted in ascending order.
        Qclip (float, default to 0): Traces lower than this quantile are clipped to this quantile value.
            Qclip = 0 means no clipping is applied. 
        th_pertmin (float, default to 1): Maximum pertentage of unmixed traces equaling to the trace minimum.
            th_pertmin = 1 means no requirement is applied. 
        epsilon (float, default to 0): The minimum value of the input traces after scaling and shifting. 
        th_residual (float, default to 0): If not zero, The redisual of unmixing should be smaller than this value.
        nbin (int, default to 1): The temporal downsampling ratio.
            nbin = 1 means temporal downsampling is not used.
        bin_option (str, can be 'downsample' (default), 'sum', or 'mean'): 
            The method of temporal downsampling. 
            'downsample' means keep one frame and discard "nbin" - 1 frames for every "nbin" frames.
            'sum' means each binned frame is the sum of continuous "nbin" frames.
            'mean' means each binned frame is the mean of continuous "nbin" frames.
        multi_alpha (bool, default to True): 
            The approach to take when "list_alpha" has multiple elements.
            True means each element will be tested and saved independently.
            False means the largest element providing non-trivial output traces 
            will be used, which can be differnt for different neurons.
        flexible_alpha (bool, default to True): Whether a flexible alpha strategy is used 
            when the smallest alpha in "list_alpha" already caused over-regularization.
            False means the final alpha is the smallest element in "list_alpha".
            True means trying to recursively divide the smallest alpha by 2 until no over-regularization exists.

    Outputs:
        traces_nmfdemix (numpy.ndarray of float, shape = (T,n)): The resulting unmixed traces. 
            Each column is the unmixed trace of a neuron.
        list_mixout (list of numpy.ndarray of float, shape = (n1,n1)): 
            Each element is the row-normalized mixing matrix for the NMF of each neuron.

    In addition to the returned variables, more outputs are stored under the folder "dir_traces".
        There are two sub-folders under this folder.
        The sub-folder "raw" stores the neuron traces and the background traces before NMF unmixing.
        The sub-folder "alpha={}" stores the same "traces_nmfdemix" and "list_mixout", 
        as well as three other quantities characterizing the NMF unmixing process:
            list_alpha_final (list of float): Each element is the final chosen alpha for the NMF of each neuron. 
                It might be one of the elements in "list_alpha", or a value smaller than the first element.
            list_MSE (list of numpy.ndarray of float, shape = (n1,)): 
                Each element is the mean squared error (NMF residual) between 
                the input traces and the NMF-reconstructed traces for the NMF of each neuron.
            list_n_iter (list of int): Each element is the number of iterations 
                to achieve NMF convergence for the NMF of each neuron.
    '''

    if not os.path.exists(dir_traces):
        os.makedirs(dir_traces) 

    start = time.time()
    # Get information about the video
    file_video = h5py.File(filename_video, 'r')
    varname = find_dataset(file_video)
    (T, Lx, Ly) = video_shape = file_video[varname].shape
    video_dtype = file_video[varname].dtype

    # Load neuron masks
    try:
        file_masks = loadmat(filename_masks)
        Masks = file_masks['FinalMasks'].transpose([2,1,0]).astype('bool')
    except:
        file_masks = h5py.File(filename_masks, 'r')
        Masks = np.array(file_masks['FinalMasks']).astype('bool')
        file_masks.close()
    (_, Lxm, Lym) = Masks.shape
    # If the shape of the masks is different from the shapes of the video,
    # zero-pad or crop the masks to fit the video shape 
    if Lxm < Lx:
        Masks = np.pad(Masks,((0,0),(0,Lx-Lxm),(0,0)),'constant', constant_values=0)
    elif Lxm > Lx:
        Masks = Masks[:,:Lx,:]
    if Lym < Ly:
        Masks = np.pad(Masks,((0,0),(0,0),(0,Ly-Lym)),'constant', constant_values=0)
    elif Lym > Ly:
        Masks = Masks[:,:,:Ly]
    masks_shape = Masks.shape

    # Determine the method to calculate traces: shared memory, memory mapping, or numpy
    if Masks.sum()*T < 7e7:
        trace_method = 'numba' # Using numba is faster for small videos
    else:
        if (sys.version_info.major+sys.version_info.minor/10)>=3.8:
            trace_method = 'shm' # "shared_memory" module is only available on python versions >= 3.8
        else:
            trace_method = 'memmap' # memory mapping is available on older python versions
        
    if trace_method == 'shm':
        # Create the shared memory object for the video
        nbytes_video = int(video_dtype.itemsize * file_video[varname].size)
        shm_video = SharedMemory(create=True, size=nbytes_video)
        video = np.frombuffer(shm_video.buf, dtype = file_video[varname].dtype)
        video[:] = file_video[varname][()].ravel()
        video = video.reshape(file_video[varname].shape)

        # Create the shared memory object for the masks
        shm_masks = SharedMemory(create=True, size=Masks.nbytes)
        FinalMasks = np.frombuffer(shm_masks.buf, dtype = 'bool')
        FinalMasks[:] = Masks.ravel()
        FinalMasks = FinalMasks.reshape(Masks.shape)
        del Masks

    elif trace_method == 'memmap':
        name_mmap = Exp_ID
        # Create the memory mapping file for the video
        fn_video = name_mmap + 'video.dat'
        fp_video = np.memmap(fn_video, dtype=video_dtype, mode='w+', shape=(T,Lx*Ly))
        for tt in range(T):
            fp_video[tt] = file_video[varname][tt].ravel()

        # Create the memory mapping file for the masks
        fn_masks = name_mmap + 'masks.dat'
        fp_masks = np.memmap(fn_masks, dtype='bool', mode='w+', shape=masks_shape)
        fp_masks[:] = Masks[:]
        del Masks

    elif trace_method == 'numba':
        video = np.array(file_video[varname])

        
    file_video.close()
    finish = time.time()
    print('Data loading time: {} s'.format(finish - start))

    # Calculate traces, background, and outside traces
    start = time.time()
    if trace_method == 'shm':
        (traces, bgtraces, outtraces, list_neighbors) = traces_bgtraces_from_masks_shm_neighbors(
            shm_video, video_dtype, video_shape, shm_masks, masks_shape, FinalMasks)
    elif trace_method == 'memmap':
        (traces, bgtraces, outtraces, list_neighbors) = traces_bgtraces_from_masks_mmap_neighbors(
            fn_video, video_dtype, video_shape, fn_masks, masks_shape, fp_masks, name_mmap)
    elif trace_method == 'numba':
        (traces, bgtraces, outtraces, list_neighbors) = traces_bgtraces_from_masks_numba_neighbors(
            video, Masks)

    # Save the raw traces into a ".mat" file under folder "dir_trace_raw".
    dir_trace_raw = os.path.join(dir_traces, "raw")
    finish = time.time()
    print('Trace calculation time: {} s'.format(finish - start))
    if not os.path.exists(dir_trace_raw):
        os.makedirs(dir_trace_raw)        
    savemat(os.path.join(dir_trace_raw, Exp_ID+".mat"), {"traces": traces, "bgtraces": bgtraces})

    if not isinstance(list_alpha, list):
        list_alpha = [list_alpha]
    else:
        list_alpha.sort()

    if multi_alpha:
        for alpha in list_alpha:
            # Apply NMF to unmix the background-subtracted traces
            start = time.time()
            traces_nmfdemix, list_mixout, list_MSE, list_final_alpha, list_n_iter = \
                use_nmfunmix(traces, bgtraces, outtraces, list_neighbors, [alpha], Qclip, \
                th_pertmin, epsilon, th_residual, nbin, bin_option, flexible_alpha)
            finish = time.time()
            print('NMF unmixing time: {} s'.format(finish - start))

            # Save the unmixed traces into a ".mat" file under folder "dir_trace_unmix".
            dir_trace_unmix = os.path.join(dir_traces, "alpha={:6.3f}".format(alpha))
            if not os.path.exists(dir_trace_unmix):
                os.makedirs(dir_trace_unmix)        
            savemat(os.path.join(dir_trace_unmix, Exp_ID+".mat"), {"traces_nmfdemix": traces_nmfdemix,\
                "list_mixout":list_mixout, "list_MSE":list_MSE, "list_final_alpha":list_final_alpha, "list_n_iter":list_n_iter})

    else:
        # Apply NMF to unmix the background-subtracted traces
        start = time.time()
        traces_nmfdemix, list_mixout, list_MSE, list_final_alpha, list_n_iter = \
            use_nmfunmix(traces, bgtraces, outtraces, list_neighbors, list_alpha, Qclip, \
            th_pertmin, epsilon, th_residual, nbin, bin_option, flexible_alpha)
        finish = time.time()
        print('Unmixing time: {} s'.format(finish - start))

        # Save the unmixed traces into a ".mat" file under folder "dir_trace_unmix".
        if len(list_alpha) > 1:
            dir_trace_unmix = os.path.join(dir_traces, "list_alpha={}".format(str(list_alpha)))
        else:
            dir_trace_unmix = os.path.join(dir_traces, "alpha={:6.3f}".format(list_alpha[0]))
        if not os.path.exists(dir_trace_unmix):
            os.makedirs(dir_trace_unmix)        
        savemat(os.path.join(dir_trace_unmix, Exp_ID+".mat"), {"traces_nmfdemix": traces_nmfdemix,\
            "list_mixout":list_mixout, "list_MSE":list_MSE, "list_final_alpha":list_final_alpha, "list_n_iter":list_n_iter})


    if trace_method == 'shm':
        # Unlink shared memory objects
        shm_video.close()
        shm_video.unlink()
        shm_masks.close()
        shm_masks.unlink()
    elif trace_method == 'memmap':
        # Delete memory mapping files
        fp_masks._mmap.close()
        del fp_masks
        os.remove(fn_masks)
        fp_video._mmap.close()
        del fp_video
        os.remove(fn_video)

    return traces_nmfdemix, list_mixout




import numpy as np


def bin_trace(trace, nbin=1, bin_option='downsample'):
    ''' Unmix the input traces in Ftmix using NMF, and obtain the unmixed traces and the mixing matrix. 
    Inputs: 
        trace (numpy.ndarray of float, shape = (T,n)): The input unbinned traces.
            Each column represents an input temporal trace.
        nbin (int, default to 1): The temporal downsampling ratio.
            nbin = 1 means temporal downsampling is not used.
        bin_option (str, can be 'downsample' (default), 'sum', or 'mean'): 
            The method of temporal downsampling. 
            'downsample' means keep one frame and discard "nbin" - 1 frames for every "nbin" frames.
            'sum' means each binned frame is the sum of continuous "nbin" frames.
            'mean' means each binned frame is the mean of continuous "nbin" frames.

    Outputs:
        trace_bin (numpy.ndarray of float, shape = (T//nbin,n)): The resulting binned traces. 
            Each column is the binned trace corresponding to each column in trace.
    '''

    if nbin==1:
        trace_bin = trace
    else:
        (T,n) = trace.shape
        trace_bin = np.zeros((T//nbin,n), dtype = trace.dtype)
        if bin_option == 'sum':
            for t in range(T//nbin):
                trace_bin[t] = trace[t*nbin:(t+1)*nbin].sum(0)
        elif bin_option == 'mean':
            for t in range(T//nbin):
                trace_bin[t] = trace[t*nbin:(t+1)*nbin].mean(0)
        elif bin_option == 'downsample':
            for t in range(T//nbin):
                trace_bin[t] = trace[t*nbin]
        else:
            raise ValueError("bin_option must be 'downsample', 'sum', or 'mean'")

    return trace_bin
    

def bin_video(video, nbin=1, bin_option='downsample'):
    ''' Unmix the input traces in Ftmix using NMF, and obtain the unmixed traces and the mixing matrix. 
    Inputs: 
        video (numpy.ndarray of float, shape = (T,Lx,Ly)): The input unbinned video.
        nbin (int, default to 1): The temporal downsampling ratio.
            nbin = 1 means temporal downsampling is not used.
        bin_option (str, can be 'downsample' (default), 'sum', or 'mean'): 
            The method of temporal downsampling. 
            'downsample' means keep one frame and discard "nbin" - 1 frames for every "nbin" frames.
            'sum' means each binned frame is the sum of continuous "nbin" frames.
            'mean' means each binned frame is the mean of continuous "nbin" frames.

    Outputs:
        video_bin (numpy.ndarray of float, shape = (T//nbin,Lx,Ly)): The resulting binned video. 
    '''

    if nbin==1:
        video_bin = video
    else:
        (T,Lx,Ly) = video.shape
        video_bin = np.zeros((T//nbin,Lx,Ly), dtype = video.dtype)
        if bin_option == 'sum':
            for t in range(T//nbin):
                video_bin[t] = video[t*nbin:(t+1)*nbin].sum(0)
        elif bin_option == 'mean':
            for t in range(T//nbin):
                video_bin[t] = video[t*nbin:(t+1)*nbin].mean(0)
        elif bin_option == 'downsample':
            for t in range(T//nbin):
                video_bin[t] = video[t*nbin]
        else:
            raise ValueError("bin_option must be 'downsample', 'sum', or 'mean'")

    return video_bin




import numpy as np
from nmfunmix_MSE import nmfunmix


def nmfunmix1(i, trace, outtrace, list_alpha=[0], th_pertmin=1, epsilon=0, \
        th_residual=0, nbin=1, bin_option='downsample', flexible_alpha=True):
    ''' Unmix the input traces in "trace" and "outtrace" using NMF, and obtain the unmixed traces and the mixing matrix. 
    Inputs: 
        i (int): The index of the neuron of interest.
        trace (numpy.ndarray of float, shape = (T,n-1)): The background-subtracted traces for neurons,
            including the neuron of interest and neighboring neurons.
            Each column represents an input temporal trace, which can be a trace of F, dF/F, or SNR.
            The first column (trace[:,0]) is the time series of the neuron of interest.
            Each of the remaining columns (trace[:,1:]), if any, should be the time series of a neighboring neuron.
        outtrace (numpy.ndarray of float, shape = (T,)): The background-subtracted trace of the outside activities.
            - the outside activities should be the average of the pixels around the neuron of interest, 
            and these pixels should not be in the masks of the neighboring neuorns included in "trace".
        list_alpha (list of float, default to [0]): A list of alpha to be tested.
            The elements should be sorted in ascending order.
        th_pertmin (float, default to 1): Maximum pertentage of unmixed traces equaling to the trace minimum.
            th_pertmin = 1 means no requirement is applied. 
        epsilon (float, default to 0): The minimum value of the input traces after scaling and shifting. 
        th_residual (float, default to 0): If not zero, The redisual of unmixing should be smaller than this value.
        nbin (int, default to 1): The temporal downsampling ratio.
            nbin = 1 means temporal downsampling is not used.
        bin_option (str, can be 'downsample' (default), 'sum', or 'mean'): 
            The method of temporal downsampling. 
            'downsample' means keep one frame and discard "nbin" - 1 frames for every "nbin" frames.
            'sum' means each binned frame is the sum of continuous "nbin" frames.
            'mean' means each binned frame is the mean of continuous "nbin" frames.
        flexible_alpha (bool, default to True): Whether a flexible alpha strategy is used 
            when the smallest alpha in "list_alpha" already caused over-regularization.
            False means the final alpha is the smallest element in "list_alpha".
            True means trying to recursively divide the smallest alpha by 2 until no over-regularization exists.

    Outputs:
        traceout (numpy.ndarray of float, shape = (T,n)): The resulting unmixed traces. 
            Each column is the unmixed trace corresponding to each column in tracein.
        mixout (numpy.ndarray of float, shape = (n,n)): The row-normalized mixing matrix.
        outtrace (numpy.ndarray of float, shape = (T,n)): Same as the input "outtrace".
        tempmixIDs (numpy.ndarray of float, binary value, shape = (n,n)):
            The assistent transformation matrix used to match raw output traces to the input traces.
        subtraces (list of int): The index of identically zero output traces.
            Idealy should be empty.
        alpha_final (float): The final chosen alpha. It might be one of the elements in "list_alpha", 
            or a value smaller than the first element.
        MSE (numpy.ndarray of float, shape = (n,)): 
            Mean squared error (NMF residual) between the input traces and the NMF-reconstructed traces.
        tracein (numpy.ndarray of float, shape = (T,n)): The input traces to NMF. 
            Each column represents an input temporal trace.
            The first column (tracein[:,0]) is the time series of the neuron of interest.
            The last column (tracein[:,-1]) should be the time series of the outside activities
            - the outside activities should be the average of the pixels around the neuron of interest, 
            and these pixels should not be in the masks of the neighboring neuorns included in "tracein".
            Each of the remaining columns (tracein[:,1:-1]), if any, should be the time series of a neighboring neuron.
        n_iter (int): Number of iterations to achieve NMF convergence.
    '''

    eps = np.finfo(np.float32).eps
    if not isinstance(list_alpha, list):
        list_alpha = [list_alpha]
    else:
        list_alpha.sort()

    tracein = np.concatenate([trace, outtrace[:,np.newaxis]], axis=1)
    noise = th_residual * np.ones(tracein.shape[1])

    # call nmfunmix
    # Gradually increase alpha until the output mixout is singular,
    # then choose the maximum alpha that provides nonsingular mixout
    for j in range(len(list_alpha)):
        (traceout, mixout, tempmixIDs, subtraces, MSE, n_iter) = nmfunmix(tracein, \
            nbin=nbin, alpha=list_alpha[j], epsilon=epsilon, bin_option=bin_option) 
        alpha_final = list_alpha[j]
        question_flag = over_regularization(traceout, eps, subtraces, th_pertmin, th_residual, MSE, noise)
        
        # question_flag == True means the alpha is too large. 
        if question_flag:
            jj = j*1
            if jj > 0: # If the alpha is too large only at some late alpha, then return to the privous alpha
                while jj > 0 and question_flag:
                    jj = jj-1
                    (traceout, mixout, tempmixIDs, subtraces, MSE, n_iter) = nmfunmix(tracein, \
                        nbin=nbin, alpha=list_alpha[jj], epsilon=epsilon, bin_option=bin_option) 
                    alpha_final = list_alpha[jj]
                    question_flag = over_regularization(traceout, eps, subtraces, th_pertmin, th_residual, MSE, noise) 
            if jj == 0 and flexible_alpha:  
                # if the first alpha already caused over-regularization, and flexible alpha strategy is used,
                # then recursively divide alpha by 2 until no over-regularization exists.
                alpha_temp = list_alpha[0]
                while question_flag and alpha_temp>1e-4: # list_alpha[0]/5: # 
                    alpha_temp = alpha_temp/2
                    (traceout, mixout, tempmixIDs, subtraces, MSE, n_iter) = nmfunmix(tracein, \
                        nbin=nbin, alpha=alpha_temp, epsilon=epsilon, bin_option=bin_option) 
                    alpha_final = alpha_temp
                    question_flag = over_regularization(traceout, eps, subtraces, th_pertmin, th_residual, MSE, noise)
            break

    print('finished neuron', i)
    return traceout, mixout, outtrace, tempmixIDs, subtraces, alpha_final, MSE, tracein, n_iter


def over_regularization(traceout, eps, subtraces, th_pertmin, th_residual, MSE, noise):
    # "pertmin" is the indicator of whether the pertentage of unmixed traces equaling to the trace minimum exceeds "th_pertmin"
    if th_pertmin < 1:
        pertmin = (np.abs(traceout - traceout.min(0))<eps).mean(0).max() > th_pertmin
    else: 
        pertmin = False
    # "noisy" is the indicator of whether the the residual exceeds "th_residual"
    if th_residual:
        # noisy = MSE[0] > noise[0]
        # noisy = MSE.mean() > noise.mean()
        # noisy = np.all(MSE > noise)
        noisy = np.any(MSE > noise)
    else:
        noisy = False
    question_flag = pertmin or subtraces.size or noisy 
    return question_flag



import numpy as np
from numba import jit, prange


@jit("void(u2[:,:],f4[:])",nopython=True,parallel=True,cache=True,fastmath=True)
def fast_mean_u2(video, trace):
    '''Calculate the means of each frame as the temporal traces
    Inputs: 
        video(numpy.ndarray of uint16, shape = (T,L)): The input video reshaped to 2D.
    Outputs:
        trace(numpy.ndarray of float32, shape = (L,)): The means of each frame.
    '''
    for tt in prange(video.shape[0]):
        trace[tt] = np.mean(video[tt, :])


@jit("void(u2[:,:],f4[:])",nopython=True,parallel=True,cache=True,fastmath=True)
def fast_median_u2(video, bgtrace):
    '''Calculate the medians of each frame as the temporal traces
    Inputs: 
        video(numpy.ndarray of uint16, shape = (T,L)): The input video reshaped to 2D.
    Outputs:
        bgtrace(numpy.ndarray of float32, shape = (L,)): The medians of each frame.
    '''
    for tt in prange(video.shape[0]):
        bgtrace[tt] = np.median(video[tt, :])


@jit("void(f4[:,:],f4[:])",nopython=True,parallel=True,cache=True,fastmath=True)
def fast_mean_f4(video, trace):
    '''Calculate the means of each frame as the temporal traces
    Inputs: 
        video(numpy.ndarray of float32, shape = (T,L)): The input video reshaped to 2D.
    Outputs:
        trace(numpy.ndarray of float32, shape = (L,)): The means of each frame.
    '''
    for tt in prange(video.shape[0]):
        trace[tt] = np.mean(video[tt, :])


@jit("void(f4[:,:],f4[:])",nopython=True,parallel=True,cache=True,fastmath=True)
def fast_median_f4(video, bgtrace):
    '''Calculate the medians of each frame as the temporal traces
    Inputs: 
        video(numpy.ndarray of float32, shape = (T,L)): The input video reshaped to 2D.
    Outputs:
        bgtrace(numpy.ndarray of float32, shape = (L,)): The medians of each frame.
    '''
    for tt in prange(video.shape[0]):
        bgtrace[tt] = np.median(video[tt, :])


def traces_bgtraces_from_masks_numba_neighbors(video, masks):
    ''' Calculate the traces of the neuron, background, and outside region for a neuron. 
    Inputs: 
        video (numpy.ndarray, shape = (T,Lx,Ly)): The input video.
        masks (numpy.ndarray of bool, shape = (n,Lx,Ly)): The binary spatial masks of all neurons.
            Each slice represents a binary mask of a neuron.

    Outputs:
        trace (numpy.ndarray of float32, shape = (T,)): The raw trace of the neuron.
        bgtrace (numpy.ndarray of float32, shape = (T,)): The raw background trace around the neuron.
        outtrace (numpy.ndarray of float32, shape = (T,)): The raw trace of all the outside activities around the neuron.
    '''

    (T, Lx, Ly) = video.shape
    (ncells, Lxm, Lym) = masks.shape

    # Find the neighboring neurons of all neurons. 
    (list_neighbors, comx, comy, area, r_bg) = find_neighbors(masks)
    [xx, yy] = np.meshgrid(np.arange(Ly), np.arange(Lx))
    xx = xx.astype('uint16')
    yy = yy.astype('uint16')

    # Reshape 3D video to 2D
    video = video.reshape((T, Lxm*Lym))

    traces = np.zeros((T, ncells), dtype='float32')
    bgtraces = np.zeros((T, ncells), dtype='float32')
    outtraces = np.zeros((T, ncells), dtype='float32')

    if video.dtype == 'uint16':
        fast_mean = fast_mean_u2
        fast_median = fast_median_u2
    elif video.dtype == 'float32':
        fast_mean = fast_mean_f4
        fast_median = fast_median_f4
    else:
        fast_mean = fast_mean_f4
        fast_median = fast_median_f4
        video = video.astype('float32')

    for nn in range(ncells):
        mask = masks[nn]
        # Expand the radius of the background mask if the neuron area is large.
        r_bg_0 = max(r_bg, np.sqrt(area[nn]/np.pi) * 2.5)
        circleout = (yy-comx[nn]) ** 2 + (xx-comy[nn]) ** 2 < r_bg_0 ** 2
        fast_median(video[:, circleout.ravel()], bgtraces[:, nn]) # Background trace
        fast_mean(video[:, mask.ravel()], traces[:, nn]) # Neuron trace

        fgmask = masks[list_neighbors[nn]].sum(0) > 0
        r_bg_large = r_bg * 0.8
        circleout = (yy-comx[nn]) ** 2 + (xx-comy[nn]) ** 2 < r_bg_large ** 2
        bgmask = np.logical_and(circleout,np.logical_not(fgmask)) # Outside mask
        bgweight = bgmask.sum()
        # Adjust the radius of the outside mask, so that the area of the outside mask is larger than half of the neuron area
        while bgweight < area[nn]/2: # == 0: # 
            r_bg_large = r_bg_large+1
            circleout = (yy-comx[nn]) ** 2 + (xx-comy[nn]) ** 2 < r_bg_large ** 2
            bgmask = np.logical_and(circleout,np.logical_not(fgmask))
            bgweight = bgmask.sum()
        fast_mean(video[:, bgmask.ravel()], outtraces[:, nn]) # Outside trace

    return traces, bgtraces, outtraces, list_neighbors


def find_neighbors(FinalMasks):
    ''' Find the neighboring neurons of all neurons. 
    Inputs: 
        FinalMasks (numpy.ndarray of bool, shape = (n,Lx,Ly)): The binary spatial masks of all neurons.
            Each slice represents a binary mask of a neuron.

    Outputs:
        list_neighbors (list of list of int): 
            Each element is a list of indeces of neighboring neurons of a neuron.
        comx (numpy.ndarray of float, shape = (n,)): The x-position of the center of each mask.
        comy (numpy.ndarray of float, shape = (n,)): The y-position of the center of each mask.
        area (numpy.ndarray of int, shape = (n,)): The area of each mask.
        r_bg (float): The radius of the background mask.
    '''

    # (n, Lx, Ly) = FinalMasks.shape
    n = FinalMasks.shape[0]
    list_neighbors = []
    area = np.zeros(n)
    comx = np.zeros(n)
    comy = np.zeros(n)
    for i in range(n):
        mask = FinalMasks[i]
        [xxs, yys] = mask.nonzero()
        area[i] = xxs.size
        comx[i] = xxs.mean()
        comy[i] = yys.mean()

    r_bg = np.sqrt(area.mean()/np.pi)*2.5
    for i in range(n):
        # The distance between each pair of neurons
        r = np.sqrt((comx[i]-comx) **2 + (comy[i]-comy) **2)
        # Neighboring neurons are defined as the neurons whose centers are 
        # less than "r_bg" away from the center of the neuron of interest. 
        neighbors = np.concatenate([np.array([i]), np.logical_and(r > 0, r < r_bg).nonzero()[0]])
        list_neighbors.append(neighbors)

    return list_neighbors, comx, comy, area, r_bg





import os
import numpy as np
import multiprocessing as mp


def mean_median_out_trace(nn, fn_video, video_dtype, video_shape, \
    fn_masks, fn_xx, fn_yy, r_bg, comx, comy, area, neighbors):
    ''' Calculate the traces of the neuron, background, and outside region for a neuron. 
    Inputs: 
        nn (int): The index of the neuron of interest.
        fn_video (str): Memory mapping file name of the input video.
        video_dtype (str): The data type of the input video.
        video_shape (tuple of int, shape = (3,)): The shape of the input video.
        fn_masks (str): Memory mapping file name of the binary spatial masks of all neurons.
        fn_xx (str): Memory mapping file name of the x position of a grid.
        fn_yy (str): Memory mapping file name of the y position of a grid.
        r_bg (float): The radius of the background mask.
        comx (numpy.ndarray of float, shape = (n,)): The x-position of the center of all masks.
        comy (numpy.ndarray of float, shape = (n,)): The y-position of the center of all masks.
        area (int): The area of the mask.
        neighbors (list of int): A list of indeces of neighboring neurons.

    Outputs:
        trace (numpy.ndarray of float, shape = (T,)): The raw trace of the neuron.
        bgtrace (numpy.ndarray of float, shape = (T,)): The raw background trace around the neuron.
        outtrace (numpy.ndarray of float, shape = (T,)): The raw trace of all the outside activities around the neuron.
    '''

    (T, Lx, Ly) = video_shape
    n = comx.size
    # Reconstruct the video, masks, and corrdinates from memory mapping files.
    fp_video = np.memmap(fn_video, dtype=video_dtype, mode='r', shape=(T,Lx*Ly))
    fp_masks = np.memmap(fn_masks, dtype='bool', mode='r', shape=(n,Lx,Ly))
    mask = fp_masks[nn]
    fp_xx = np.memmap(fn_xx, dtype='uint16', mode='r', shape=(Lx,Ly))
    fp_yy = np.memmap(fn_yy, dtype='uint16', mode='r', shape=(Lx,Ly))

    # Expand the radius of the background mask if the neuron area is large.
    r_bg_0 = max(r_bg, np.sqrt(area/np.pi) * 2.5)
    circleout = (fp_yy-comx[nn]) ** 2 + (fp_xx-comy[nn]) ** 2 < r_bg_0 ** 2 # Background mask
    bgtrace = np.median(fp_video[:, circleout.ravel()], 1) # Background trace
    trace = fp_video[:, mask.ravel()].mean(1) # Neuron trace

    fgmask = fp_masks[neighbors].sum(0) > 0
    r_bg_large = r_bg * 0.8
    circleout = (fp_yy-comx[nn]) ** 2 + (fp_xx-comy[nn]) ** 2 < r_bg_large ** 2
    bgmask = np.logical_and(circleout,np.logical_not(fgmask)) # Outside mask
    bgweight = bgmask.sum() # Area of outside mask
    # Adjust the radius of the outside mask, so that the area of the outside mask is larger than half of the neuron area
    while bgweight < area/2: # == 0: # 
        r_bg_large = r_bg_large+1
        circleout = (fp_yy-comx[nn]) ** 2 + (fp_xx-comy[nn]) ** 2 < r_bg_large ** 2
        bgmask = np.logical_and(circleout,np.logical_not(fgmask))
        bgweight = bgmask.sum()
    outtrace = fp_video[:, bgmask.ravel()].mean(1) # Outside trace

    return trace, bgtrace, outtrace


def traces_bgtraces_from_masks_mmap_neighbors(fn_video, video_dtype, video_shape, \
        fn_masks, masks_shape, fp_masks, name_mmap):
    ''' Calculate the traces of the neuron, background, and outside region for all neurons. 
    Inputs: 
        fn_video (str): Memory mapping file name of the input video.
        video_dtype (str): The data type of the input video.
        video_shape (tuple of int, shape = (3,)): The shape of the input video.
        fn_masks (str): Memory mapping file name of the binary spatial masks of all neurons.
        masks_shape (tuple of int, shape = (3,)): The shape of the neuron masks.
        fp_masks (numpy.ndarray of bool, shape = (n,Lx,Ly)): The binary spatial masks of all neurons.
            Each slice represents a binary mask of a neuron.
        name_mmap (str): Identification section of the memory mapping file name.

    Outputs:
        traces (numpy.ndarray of float, shape = (T,n)): The raw traces of all neurons.
            Each column represents a trace of a neuron.
        bgtraces (numpy.ndarray of float, shape = (T,n)): All the raw background traces.
            Each column represents a background trace corresponding to a neuron.
        outtraces (numpy.ndarray of float, shape = (T,n)): The raw traces of all the outside activities.
            Each column represents an outside trace corresponding to a neuron.
        list_neighbors (list of list of int): 
            Each element is a list of indeces of neighboring neurons of a neuron.
    '''

    (ncells, Lx, Ly) = masks_shape
    # Find the neighboring neurons of all neurons. 
    (list_neighbors, comx, comy, area, r_bg) = find_neighbors(fp_masks)

    # Create memory mapping files for x and y coordinates
    [xx, yy] = np.meshgrid(np.arange(Ly), np.arange(Lx))
    xx = xx.astype('uint16')
    fn_xx = name_mmap + 'xx.dat'
    fp_xx = np.memmap(fn_xx, dtype='uint16', mode='w+', shape=(Lx,Ly))
    fp_xx[:] = xx[:]
    yy = yy.astype('uint16')
    fn_yy = name_mmap + 'yy.dat'
    fp_yy = np.memmap(fn_yy, dtype='uint16', mode='w+', shape=(Lx,Ly))
    fp_yy[:] = yy[:]

    # results = []
    # for nn in range(ncells):
    #     results.append(mean_median_out_trace(nn, shm_video, video_dtype, video_shape, \
    #         shm_masks, shm_xx, shm_yy, r_bg, comx, comy, area[nn], list_neighbors[nn]))

    # Calculate the traces of the neuron, background, and outside region for each neuron. 
    p = mp.Pool(mp.cpu_count())
    results = p.starmap(mean_median_out_trace, [(nn, fn_video, video_dtype, video_shape, \
        fn_masks, fn_xx, fn_yy, r_bg, comx, comy, area[nn], list_neighbors[nn]) for nn in range(ncells)], chunksize=1)
    p.close()

    traces = np.vstack([x[0] for x in results]).T
    bgtraces = np.vstack([x[1] for x in results]).T
    outtraces = np.array([x[2] for x in results]).T

    # Delete memory mapping files
    fp_xx._mmap.close()
    del fp_xx
    os.remove(fn_xx)
    fp_yy._mmap.close()
    del fp_yy
    os.remove(fn_yy)

    return traces, bgtraces, outtraces, list_neighbors


def find_neighbors(FinalMasks):
    ''' Find the neighboring neurons of all neurons. 
    Inputs: 
        FinalMasks (numpy.ndarray of bool, shape = (n,Lx,Ly)): The binary spatial masks of all neurons.
            Each slice represents a binary mask of a neuron.

    Outputs:
        list_neighbors (list of list of int): 
            Each element is a list of indeces of neighboring neurons of a neuron.
        comx (numpy.ndarray of float, shape = (n,)): The x-position of the center of each mask.
        comy (numpy.ndarray of float, shape = (n,)): The y-position of the center of each mask.
        area (numpy.ndarray of int, shape = (n,)): The area of each mask.
        r_bg (float): The radius of the background mask.
    '''

    # (n, Lx, Ly) = FinalMasks.shape
    n = FinalMasks.shape[0]
    list_neighbors = []
    area = np.zeros(n, dtype = 'uint32')
    comx = np.zeros(n)
    comy = np.zeros(n)
    for i in range(n):
        mask = FinalMasks[i]
        [xxs, yys] = mask.nonzero()
        area[i] = xxs.size
        comx[i] = xxs.mean()
        comy[i] = yys.mean()

    r_bg = np.sqrt(area.mean()/np.pi)*2.5
    for i in range(n):
        # The distance between each pair of neurons
        r = np.sqrt((comx[i]-comx) **2 + (comy[i]-comy) **2)
        # Neighboring neurons are defined as the neurons whose centers are 
        # less than "r_bg" away from the center of the neuron of interest. 
        neighbors = np.concatenate([np.array([i]), np.logical_and(r > 0, r < r_bg).nonzero()[0]])
        list_neighbors.append(neighbors)

    return list_neighbors, comx, comy, area, r_bg




#!/usr/bin/env python
# coding=utf-8

from setuptools import setup, find_packages

setup(
    name='tuncat',
    version='1.0',
    description=(
        'Temporal Unmixing of Calcium Traces (TUnCaT) is an automatic algorithm to decontaminate false transients from temporal traces generated from fluorescent calcium imaging videos. '
    ),
    long_description=open('README.md').read(),
    # long_description=long_description,
    long_description_content_type="text/markdown",
    author='Yijun Bao',
    author_email='yijun.bao@duke.edu',
    maintainer='Yijun Bao',
    maintainer_email='yijun.bao@duke.edu',
    license='the GNU License, Version 2.0',
    packages=find_packages(), # ['tuncat'], # 
    platforms=["all"],
    url='https://github.com/YijunBao/TUnCaT',
    project_urls={
        "Source Code": "https://github.com/YijunBao/TUnCaT",
        "Citation": "https://doi.org/10.5281/zenodo.5764576",
        "Paper": "https://doi.org/10.3389/fnins.2021.797421"
    },
    classifiers=[
        'Operating System :: OS Independent',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering :: Bio-Informatics'
    ],
    install_requires=[
        'numpy',
        'scipy',
        'scikit-learn',
        'numba',
        'h5py'
    ],
    python_requires='>=3'
    )